from brownie import chain
from ypricemagic import magic
from ypricemagic.constants import weth
from ypricemagic.exceptions import UnsupportedNetwork
from ypricemagic.networks import Network
from ypricemagic.utils.raw_calls import raw_call


class wstEth:
    def __init__(self) -> None:
        try: self.address = {
            Network.Mainnet: '0x7f39C581F595B53c5cb19bD0b3f8dA6c935E2Ca0'
            }[chain.id]
        except KeyError: self.address = None
    
    def get_price(self, block=None):
        share_price = raw_call(self.address, 'stEthPerToken()', output='int', block=block) / 1e18
        return share_price * magic.get_price(weth, block)

wsteth = wstEth()

def is_wsteth(address):
    try: return address == wsteth.address
    except UnsupportedNetwork: return False




from collections import defaultdict
from itertools import count, product
from operator import itemgetter
from typing import List

import requests
from brownie import chain, web3
from eth_abi.exceptions import InsufficientDataBytes
from web3.exceptions import CannotHandleRequest
from ypricemagic.utils.contracts import Contract, contract_creation_block
from ypricemagic.utils.raw_calls import _decimals, _totalSupply

from multicall import Call, Multicall

# the address doesn't matter, it just needs to have code
code = " "
multicall, multicall2 = None, None
if chain.id == 1:
    multicall2 = Contract("0x5BA1e12693Dc8F9c48aAD8770482f4739bEeD696")
if chain.id == 56:
    multicall2 = Contract("0xfF6FD90A470Aaa0c1B8A54681746b07AcdFedc9B")
if chain.id == 137:
    multicall2 = Contract("0xc8E51042792d7405184DfCa245F2d27B94D013b6")
if chain.id == 250:
    multicall2 = Contract("0xBAD2B082e2212DE4B065F636CA4e5e0717623d18")

multicall_deploy_block = contract_creation_block(multicall2.address)

def multicall_same_func_no_input(addresses: list, method: str, apply_func=None, block=None):
    calls = [Call(address, [method], [[address,apply_func]]) for address in addresses]
    return [result for result in Multicall(calls, block_id=block, _w3=web3)().values()]

def multicall_same_func_different_contracts_same_input(addresses: list, method: str, input=None, apply_func=None, block=None):
    assert input
    calls = [Call(address, [method, input], [[address,apply_func]]) for address in addresses]
    return [result for result in Multicall(calls, block_id=block, _w3=web3)().values()]

def multicall_same_func_same_contract_different_inputs(address: list, method: str, inputs: List=None, apply_func=None, block=None):
    assert inputs
    calls = [Call(address, [method, input], [[input,apply_func]]) for input in inputs]
    return [result for result in Multicall(calls, block_id=block, _w3=web3)().values()]

def multicall_decimals(addresses: List[str], block=None):
    try: return multicall_same_func_no_input(addresses, 'decimals()(uint256)', block=block)
    except ValueError as e:
        if 'execution reverted' in str(e): return [_decimals(address,block,return_None_on_failure=True) for address in addresses]
        else: raise
    except (CannotHandleRequest,InsufficientDataBytes): return [_decimals(address,block,return_None_on_failure=True) for address in addresses] 

def multicall_totalSupply(addresses: List[str], block=None):
    try: return multicall_same_func_no_input(addresses, 'totalSupply()(uint256)', block=block)
    except (CannotHandleRequest,InsufficientDataBytes): return [_totalSupply(address,block) for address in addresses] 

def multicall_balanceOf(token_addresses: List[str], hodler_address: str, block=None):
    return multicall_same_func_different_contracts_same_input(token_addresses, 'balanceOf(address)(uint)', input=hodler_address)

def fetch_multicall(*calls, block=None):
    # https://github.com/makerdao/multicall
    multicall_input = []
    fn_list = []
    decoded = []

    for contract, fn_name, *fn_inputs in calls:
        fn = getattr(contract, fn_name)

        # check that there aren't multiple functions with the same name
        if hasattr(fn, "_get_fn_from_args"):
            fn = fn._get_fn_from_args(fn_inputs)

        fn_list.append(fn)
        multicall_input.append((contract, fn.encode_input(*fn_inputs)))

    if isinstance(block, int) and block < multicall_deploy_block:
        # use state override to resurrect the contract prior to deployment
        data = multicall2.tryAggregate.encode_input(False, multicall_input)
        call = web3.eth.call(
            {'to': str(multicall2), 'data': data},
            block or 'latest',
            {str(multicall2): {'code': f'0x{multicall2.bytecode}'}},
        )
        result = multicall2.tryAggregate.decode_output(call)
    else:
        result = multicall2.tryAggregate.call(
            False, multicall_input, block_identifier=block or 'latest'
        )

    for fn, (ok, data) in zip(fn_list, result):
        try:
            assert ok, "call failed"
            decoded.append(fn.decode_output(data))
        except (AssertionError, InsufficientDataBytes):
            decoded.append(None)

    return decoded


def multicall_matrix(contracts, params, block="latest"):
    matrix = list(product(contracts, params))
    calls = [[contract, param] for contract, param in matrix]

    results = fetch_multicall(*calls, block=block)

    output = defaultdict(dict)
    for (contract, param), value in zip(matrix, results):
        output[contract][param] = value

    return dict(output)


def batch_call(calls):
    """
    Similar interface but block height as last param. Uses JSON-RPC batch.
    [[contract, 'func', arg, block_identifier]]
    """
    jsonrpc_batch = []
    fn_list = []
    ids = count()

    for contract, fn_name, *fn_inputs, block in calls:
        fn = getattr(contract, fn_name)
        if hasattr(fn, "_get_fn_from_args"):
            fn = fn._get_fn_from_args(fn_inputs)
        fn_list.append(fn)

        jsonrpc_batch.append(
            {
                'jsonrpc': '2.0',
                'id': next(ids),
                'method': 'eth_call',
                'params': [
                    {'to': str(contract), 'data': fn.encode_input(*fn_inputs)},
                    block,
                ],
            }
        )

    response = requests.post(web3.provider.endpoint_uri, json=jsonrpc_batch).json()
    return [
        fn.decode_output(res['result'])
        for res in sorted(response, key=itemgetter('id'))
    ]





import logging
from brownie import chain
from cachetools.func import ttl_cache
from ypricemagic.constants import usdc, weth, sushi
from ypricemagic.price_modules.uniswap.protocols import ROUTER_TO_FACTORY, special_paths
from ypricemagic.networks import Network
from ypricemagic.utils.contracts import Contract
from ypricemagic.constants import STABLECOINS
from ypricemagic.utils.raw_calls import _decimals

logger = logging.getLogger(__name__)

class UniswapRouterV2:
    def __init__(self, router_address) -> None:
        self.address = router_address
        self.contract = Contract(self.address)
        self.factory = ROUTER_TO_FACTORY[self.address]
        self.special_paths = special_paths(self.address)

    @ttl_cache(ttl=36000)
    def get_price(self, token_in, token_out=usdc, block=None, paired_against=weth):
        """
        Calculate a price based on Uniswap Router quote for selling one `token_in`.
        Always uses intermediate WETH pair if `[token_in,weth,token_out]` swap path available.
        """

        token_in, token_out = str(token_in), str(token_out)

        if chain.id == Network.BinanceSmartChain and token_out == usdc:
            busd = Contract("0xe9e7CEA3DedcA5984780Bafc599bD69ADd087D56")
            token_out = busd

        amount_in = 10 ** _decimals(token_in,block)

        if str(token_in) in STABLECOINS: return 1

        path = self.path_selector(token_in, token_out, paired_against)
        fees = 0.997 ** (len(path) - 1)
        logger.warn(f'path: {path}')
        quote = self.contract.getAmountsOut(amount_in, path, block_identifier=block)
        amount_out = quote[-1] / 10 ** _decimals(str(path[-1]),block)
        return amount_out / fees

    def path_selector(self, token_in, token_out, paired_against):

        if str(paired_against) in STABLECOINS and str(token_out) in STABLECOINS:            path = [token_in, paired_against]
        elif weth in (token_in, token_out):                                                 path = [token_in, token_out]
        elif paired_against == sushi and token_out != sushi:                                path = [token_in,sushi,weth,token_out]
        elif str(token_in) in self.special_paths and str(token_out) in STABLECOINS:  path = self.special_paths[str(token_in)]

        elif chain.id == Network.BinanceSmartChain:
            from ypricemagic.constants import cake, wbnb
            if wbnb in (token_in, token_out):                                               path = [token_in, token_out]
            elif cake in (token_in, token_out):                                             path = [token_in, token_out]
            else:                                                                           path = [token_in,wbnb,token_out]
        elif chain.id == Network.Polygon:
            from ypricemagic.constants import wmatic
            if wmatic in (token_in, token_out):                                             path = [token_in, token_out]
            else:                                                                           path = [token_in,wmatic,token_out]
        elif chain.id == Network.Fantom:
            from ypricemagic.constants import wftm
            if wftm in (token_in, token_out):                                               path = [token_in, token_out]
            else:                                                                           path = [token_in, wftm, token_out]

        else:                                                                               path = [token_in, weth, token_out]

        return path
    